require 'dry-container'
require 'edgycircle_toolbox/cqrs/message_bus'
require 'edgycircle_toolbox/cqrs/command_result'
require 'edgycircle_toolbox/cqrs/command_builder'
require 'edgycircle_toolbox/sonapi/validation_error'

module EdgycircleToolbox
  module CQRS
    class CommandBus
      def initialize
        @container = Dry::Container.new
        @command_builder = CommandBuilder.new
      end

      def register(command, handler)
        command_builder.register(command)

        container.namespace(:handlers) do
          register(command, handler)
        end
      end

      def decorate_command(command_class, decorator_class)
        command_builder.decorate(command_class, decorator_class)
      end

      def submit(command)
        result = container.resolve("handlers.#{command.class.name}").new.call(command)
        result.events.each { |event| MessageBus.publish(event) }
        result
      end

      def build(original_parameters, context = nil)
        result = CommandResult.new
        validation = command_builder.validate(original_parameters)

        if validation.failure?
          build_validation_errors(validation).each { |error| result.add_error(error) }
        else
          validated_parameters = validation.output
          command = command_builder.build(validated_parameters, context)
          result.set_command(command)
        end

        result
      end

      private
      attr_reader :container, :command_builder

      def command_class_for(name)
        container.resolve("commands.#{name}")
      end

      def build_validation_errors(validation)
        validation.messages(full: true).map do |attribute, messages|
          messages.map { |message| ValidationError.new(attribute, message) }
        end.flatten
      end
    end
  end
end
